import pytest
from pydantic import BaseModel, ValidationError, validator
from hydrothings.components.sensors.utils import metadata_validator


@pytest.mark.parametrize('metadata_value, encoding_value', [
    ('http://www.example.com/test.pdf', 'application/pdf'),
    ('http://www.example.com/test.xml', 'http://www.opengis.net/doc/IS/SensorML/2.0'),
    ('http://www.example.com/test.html', 'text/html'),
])
def test_valid_metadata_values(metadata_value, encoding_value):
    validated_value = metadata_validator(
        value=metadata_value,
        values={
            'encoding_type': encoding_value
        }
    )
    assert validated_value == metadata_value


@pytest.mark.parametrize('metadata_value, encoding_value', [
    ('http://www.example.com/test.xml', 'application/pdf'),
    ('http://www.example.com/test.html', 'application/pdf'),
    ('http://www.example.com/test.pdf', 'http://www.opengis.net/doc/IS/SensorML/2.0'),
    ('http://www.example.com/test.html', 'http://www.opengis.net/doc/IS/SensorML/2.0'),
    ('http://www.example.com/test.pdf', 'text/html'),
    ('http://www.example.com/test.xml', 'text/html'),
    ('not/a/valid/url.pdf', 'application/pdf'),
    ('not/a/valid/url.xml', 'http://www.opengis.net/doc/IS/SensorML/2.0'),
    ('not/a/valid/url.html', 'text/html')
])
def test_valid_metadata_values(metadata_value, encoding_value):
    with pytest.raises(ValueError):
        metadata_validator(
            value=metadata_value,
            values={
                'encoding_type': encoding_value
            }
        )



import pytest
from pydantic import BaseModel, ValidationError
from hydrothings.validators import allow_partial


@pytest.fixture
def thing_model():

    @allow_partial
    class Thing(BaseModel):
        id: int
        name: str
        description: str = None

    return Thing


@pytest.mark.parametrize('data_values', [
    ({}),
    ({'id': 1}),
    ({'name': 'test'}),
    ({'id': 1, 'name': 'test'}),
    ({'name': 'test', 'description': 'This is a test thing.'})
])
def test_optional_fields(thing_model, data_values):
    thing = thing_model(**data_values)
    assert thing.dict(exclude_unset=True) == data_values


@pytest.mark.parametrize('data_values', [
    ({'id': None}),
    ({'id': 1, 'name': None})
])
def test_exception_on_null(thing_model, data_values):
    with pytest.raises(ValidationError):
        thing_model(**data_values)



from django.urls import path
from hydrothings import SensorThingsAPI


# hydrothings_api_1_1 = SensorThingsAPI(
#     version='1.1'
# )
#
# hydrothings_api_1_0 = SensorThingsAPI(
#     version='1.0'
# )
#
# urlpatterns = [
#     path('v1.0/', hydrothings_api_1_0.urls),
#     path('v1.1/', hydrothings_api_1_1.urls)
# ]



import re
import hydrothings.schemas as core_schemas
from odata_query.grammar import ODataParser, ODataLexer
from pydantic import HttpUrl
from typing import Literal, Union, List
from requests import Response
from hydrothings import settings


def lookup_component(
        input_value: str,
        input_type: Literal['snake_singular', 'snake_plural', 'camel_singular', 'camel_plural'],
        output_type: Literal['snake_singular', 'snake_plural', 'camel_singular', 'camel_plural']
) -> str:
    """
    Accepts a component name and type and returns an alternate form of the component name.

    Parameters
    ----------
    input_value : str
        The name of the component to look up.
    input_type: str
        The type of the component to lookup.
    output_type : str
        The type of the component to return.

    Returns
    -------
    str
        The matching component name.
    """

    st_components = [
        {
            'snake_singular': re.sub(r'(?<!^)(?=[A-Z])', '_', capability['SINGULAR_NAME']).lower(),
            'snake_plural': re.sub(r'(?<!^)(?=[A-Z])', '_', capability['NAME']).lower(),
            'camel_singular': capability['SINGULAR_NAME'],
            'camel_plural': capability['NAME']
        } for capability in settings.ST_CAPABILITIES
    ]

    return next((c[output_type] for c in st_components if c[input_type] == input_value))


def generate_response_codes(method: str, response_schema=None) -> dict:
    """
    Generates a dictionary of response codes for various request types.

    Parameters
    ----------
    method : str
        The SensorThings method the response is for.
    response_schema : class
        An alternate response schema to attach to the API.

    Returns
    -------
    dict
        A dictionary of response codes that can be returned for the given method.
    """

    if method == 'list':
        response_codes = {
            200: response_schema
        }
    elif method == 'get':
        response_codes = {
            200: response_schema,
            403: core_schemas.PermissionDenied,
            404: core_schemas.EntityNotFound
        }
    elif method == 'create':
        response_codes = {
            201: Union[None, List[HttpUrl]],
            403: core_schemas.PermissionDenied
        }
    elif method == 'update':
        response_codes = {
            204: None,
            403: core_schemas.PermissionDenied,
            404: core_schemas.EntityNotFound
        }
    elif method == 'delete':
        response_codes = {
            204: None,
            403: core_schemas.PermissionDenied,
            404: core_schemas.EntityNotFound
        }
    else:
        response_codes = {}

    return response_codes


def entities_or_404(response):
    """"""

    if isinstance(response, Response):
        return response.status_code, response.content
    else:
        return 200, response


def entity_or_404(response, entity_id):
    """"""

    if isinstance(response, Response):
        return response.status_code, response.content
    elif response:
        return 200, response
    else:
        return 404, {'message': f'Record with ID {entity_id} does not exist.'}


def parse_query_params(
        query_params: dict,
        entity_chain: Union[List[tuple], None] = None,
        sort_datastream: bool = False
) -> dict:
    """
    Parses OData query parameters.

    This function converts OData query parameters in their string format to a dictionary of parameter objects that can
    be used to generate appropriate database queries.

    Parameters
    ----------
    query_params : dict
        A dictionary containing the raw string values of all query parameters passed with an API request.
    entity_chain : list
        A list of component/entity_id pairs representing a nested SensorThings request.
    sort_datastream : bool
        Adds a sort by datastream ID parameter if True (used in Observations requests to group by datastreams).

    Returns
    -------
    dict
        A dictionary containing all parsed query parameters.
    """

    lexer = ODataLexer()
    parser = ODataParser()

    if entity_chain:
        if query_params.get('filters'):
            query_params['filters'] += f' and {entity_chain[-1][0]}/id eq {entity_chain[-1][1]}'
        else:
            query_params['filters'] = f'{entity_chain[-1][0]}/id eq {entity_chain[-1][1]}'

    if query_params.get('filters'):
        query_params['filters'] = parser.parse(lexer.tokenize(query_params['filters']))

    if sort_datastream is True:
        if query_params.get('order_by'):
            query_params['order_by'] += ', Datastreams/id'
        else:
            query_params['order_by'] = 'Datastreams/id'

    if query_params.get('order_by'):
        query_params['order_by'] = [
            {
                'field': order_field.strip().split(' ')[0],
                'direction': 'desc' if order_field.strip().endswith('desc') else 'asc'
            } for order_field in query_params['order_by'].split(',')
        ]

    return query_params



import pytest
from django.test.client import RequestFactory
from django.urls.exceptions import Http404
from hydrothings.middleware import SensorThingsMiddleware


@pytest.fixture
def middleware():
    return SensorThingsMiddleware(lambda request: None)


@pytest.fixture
def request_factory():
    return RequestFactory()


@pytest.fixture
def service_base_url():
    return '/test-app/v1.1'


success_test_parameters = {
    'simple_url': [
        ['Things', 'Thing', []], ['Locations', 'Location', []], ['HistoricalLocations', 'HistoricalLocation', []],
        ['Sensors', 'Sensor', []], ['ObservedProperties', 'ObservedProperty', []], ['Datastreams', 'Datastream', []],
        ['Observations', 'Observation', []], ['FeaturesOfInterest', 'FeatureOfInterest', []]
    ],
    'chained_collection_url': [
        ['Things(1)/Datastreams', 'Datastream', [('Thing', '1')]],
        ['Things(2)/Locations', 'Location', [('Thing', '2')]],
        ['Datastreams(3)/Observations', 'Observation', [('Datastream', '3')]],
        ['Locations(4)/Things', 'Thing', [('Location', '4')]],
    ],
    'chained_implicit_url': [
        ['Datastreams(1)/Sensor', 'Sensor', [('Datastream', '1')]],
        ['Datastreams(2)/Thing', 'Thing', [('Datastream', '2')]],
        ['Datastreams(3)/ObservedProperty', 'ObservedProperty', [('Datastream', '3')]],
        ['Observations(4)/Datastream', 'Datastream', [('Observation', '4')]],
    ],
    'address_to_property_url': [
        ['Things(1)/name', 'Thing', []],
        ['Datastreams(2)/description', 'Datastream', []]
    ],
    'address_to_value_url': [
        ['Things(1)/name/$value', 'Thing', []],
        ['Datastreams(2)/description/$value', 'Datastream', []]
    ],
    'address_to_ref_url': [
        ['Things(1)/name/$ref', 'Thing', []],
        ['Datastreams(2)/description/$ref', 'Datastream', []]
    ]
}

failure_test_parameters = {
    'chained_unrelated_collection_url': [
        'Things(1)/Observations',
        'ObservedProperties(2)/Things'
    ],
    'chained_unrelated_implicit_url': [
        'Things(1)/Sensor',
        'Datastreams(2)/Observation'
    ],
    'address_to_nonexistent_property_url': [
        'Things(1)/color'
    ],
    'address_to_value_and_ref_url': [
        'Things(1)/name/$value/$ref'
    ]
}


@pytest.mark.parametrize('parameter_group, url, component, entity_chain', [
    tuple([parameter_group] + parameter) for parameter_group, parameters
    in success_test_parameters.items() for parameter in parameters
])
def test_successful_url_resolution_cases(
        parameter_group, url, component, entity_chain, service_base_url, middleware, request_factory
):
    request_url = f'{service_base_url}/{url}'
    request = request_factory.get(request_url)
    middleware.process_request(request)
    assert request.component == component
    assert request.entity_chain == entity_chain


@pytest.mark.parametrize('parameter_group, url', [
    (parameter_group, parameter,) for parameter_group, parameters
    in failure_test_parameters.items() for parameter in parameters
])
def test_failing_url_resolution_cases(
        parameter_group, url, service_base_url, middleware, request_factory
):
    request_url = f'{service_base_url}/{url}'
    request = request_factory.get(request_url)
    with pytest.raises(Http404):
        middleware.process_request(request)



import pytest
from pydantic import BaseModel
from hydrothings.extras.iso_types import ISOTime


@pytest.fixture
def time_model():

    class TimeModel(BaseModel):
        iso_time: ISOTime

    return TimeModel


@pytest.mark.parametrize('iso_time_value', [
    '2023-01-01T11:11:11+00:00', '2023-01-01T11:11:11'
])
def test_iso_time_valid_value(time_model, iso_time_value):
    time_instance = time_model(iso_time=iso_time_value)
    assert time_instance.iso_time == iso_time_value


@pytest.mark.parametrize('iso_time_value', [
    123, None, 'Jan 1, 2023', '2023-13-01T11:11:11'
])
def test_iso_time_invalid_value(time_model, iso_time_value):
    with pytest.raises(ValueError):
        time_model(iso_time=iso_time_value)



from abc import ABCMeta, abstractmethod
from uuid import UUID
from pydantic.fields import SHAPE_LIST
from typing import Union, Tuple, List
from django.http import HttpRequest
from hydrothings import components as component_schemas
from hydrothings import settings
from hydrothings.schemas import BasePostBody, BasePatchBody


class SensorThingsAbstractEngine(metaclass=ABCMeta):
    """
    Abstract base class for a SensorThings engine.

    A SensorThings API must be linked to a SensorThings engine to perform database operations. This class defines all
    required methods for an engine class and should be inherited by any SensorThings engine used by an API.
    """

    scheme: str
    host: str
    path: str
    version: str
    component: str
    component_path: str

    def get_ref(self, entity_id: Union[str, None] = None, related_component: Union[str, None] = None) -> str:
        """
        Builds a reference URL for a given entity.

        Parameters
        ----------
        entity_id : str
            The ID of the entity.
        related_component : str
            The related component to be appended to the ref URL.

        Returns
        -------
        str
            The entity's reference URL.
        """

        ref_url = f'{self.scheme}://{self.host}{self.path}'

        if entity_id is not None:
            ref_url = f'{ref_url}({entity_id})'

        if related_component is not None:
            ref_url = f'{ref_url}/{related_component}'

        return ref_url

    def get_related_components(self) -> dict:
        """
        Get all components related to this component.

        Returns
        -------
        dict
            The component's related components.
        """

        return {
            name: [
                component for component
                in settings.ST_CAPABILITIES
                if component['SINGULAR_NAME'] == field.type_.__name__
            ][0]['NAME'] if field.shape == SHAPE_LIST
            else field.type_.__name__
            for name, field in getattr(component_schemas, f'{self.component}Relations').__fields__.items()
        }

    def build_related_links(self, entity: dict, is_collection: bool = False) -> dict:
        """
        Creates SensorThings links to related components.

        Parameters
        ----------
        entity : dict
            The entity response object dictionary.
        is_collection : bool
            Includes the entity ID in the reference URL if True, and omits it if False.

        Returns
        -------
        dict
            The entity response object dictionary with related links included.
        """

        return dict(
            entity,
            **{
                f'{name}_link': self.get_ref(
                    entity['id'] if is_collection is True else None,
                    related_component
                ) for name, related_component in self.get_related_components().items()
            }
        )

    def build_self_links(self, entity: dict, is_collection: bool = False) -> dict:
        """
        Creates SensorThings self-referential link.

        Parameters
        ----------
        entity : dict
            The entity response object dictionary.
        is_collection : bool
            Includes the entity ID in the reference URL if True, and omits it if False.

        Returns
        -------
        dict
            The entity response object dictionary with the self link included.
        """

        return dict(
            entity,
            **{
                'self_link': self.get_ref(entity['id'] if is_collection is True else None)
            }
        )

    @abstractmethod
    def resolve_entity_id_chain(self, entity_chain: List[Tuple[str, Union[UUID, int, str]]]) -> bool:
        """
        Abstract method for resolving an entity chain passed to the API.

        SensorThings supports nested references to entities in URL paths. The Django HydroThings middleware will check
        the nested components to ensure each one is related to its parent and build a list of IDs that need to be
        verified by this method.

        Parameters
        ----------
        entity_chain : list
            A list of tuples representing the entity chain where the first object in the tuple is a string representing
            the component name, and the second object in the tuple is a string, int, or UUID representing the entity ID.

        Returns
        -------
        bool
            Returns True if the entity chain could be fully resolved, and False if any part of the entity chain could
            not be found or resolved.
        """

        pass

    @abstractmethod
    def list(
            self,
            filters,
            count,
            order_by,
            skip,
            top,
            select,
            expand
    ) -> dict:
        """
        Abstract method for handling GET collection requests.

        This method should return a dictionary representing a collection of entities for this component and apply all
        the given query parameters.

        Parameters
        ----------
        filters

        count : bool
            If True, the pre-pagination result count should be included in the response. If False, the count should be
            omitted.
        order_by

        skip : int
            An integer representing the number of records to skip in the response.
        top : int
            An integer representing the number of records the response should be limited to.
        select
            Represents a subset of this component's fields to include in the response.
        expand
            Represents related components whose fields should be included in the response.

        Returns
        -------
        dict
            A dictionary object representing the SensorThings GET collection response.
        """

        pass

    @abstractmethod
    def get(
            self,
            entity_id: str
    ) -> dict:
        """
        Abstract method for handling GET entity requests.

        This method should return a dictionary representing an entity with the given entity ID.

        Parameters
        ----------
        entity_id : str
            The ID of the entity to be returned.

        Returns
        -------
        dict
            A dictionary object representing the SensorThings GET entity response.
        """

        pass

    @abstractmethod
    def create(
            self,
            entity_body: BasePostBody,
    ) -> str:
        """
        Abstract method for handling POST entity requests.

        This method should create a new entity and return the ID of the created entity.

        Parameters
        ----------
        entity_body : BasePostBody
            A dictionary object containing the attributes of the entity that will be created.

        Returns
        -------
        str
            The ID of the newly created entity.
        """

        pass

    @abstractmethod
    def update(
            self,
            entity_id: str,
            entity_body: BasePatchBody
    ) -> str:
        """
        Abstract method for handling PATCH entity requests.

        This method should update an existing entity with the attributes included in the entity_body and return the ID
        of the updated entity.

        Parameters
        ----------
        entity_id : str
            The ID of the entity to be updated.
        entity_body : BasePatchBody
            A dictionary object containing the attributes of the entity that will be updated.

        Returns
        -------
        str
            The ID of the updated entity.
        """

        pass

    @abstractmethod
    def delete(
            self,
            entity_id
    ) -> None:
        """
        Abstract method for handling DELETE entity requests.

        This method should delete an existing entity with the given entity_id.

        Parameters
        ----------
        entity_id : str
            The ID of the entity t obe deleted.

        Returns
        -------
        None
        """

        pass


class SensorThingsRequest(HttpRequest):
    """
    The SensorThings request class.

    This class extends Django's HttpRequest class to include an engine class, component name, component path, and an
    entity_chain tuple. These attributes should be added to the request in the SensorThings middleware before calling a
    view function.
    """

    engine: SensorThingsAbstractEngine
    auth: str
    component: str
    component_path: List[str]
    entity_chain: List[Tuple[str, Union[UUID, int, str]]]



import pytest
from pydantic import BaseModel, ValidationError, validator
from hydrothings.validators import whitespace_to_none


@pytest.fixture
def thing_model():

    class Thing(BaseModel):
        name: str
        description: str = None

        _whitespace_validator = validator('*', allow_reuse=True, pre=True)(whitespace_to_none)

    return Thing


@pytest.mark.parametrize('data_values, description_model_value', [
    ({'name': 'test', 'description': ''}, None),
    ({'name': 'test', 'description': '  '}, None),
    ({'name': 'test', 'description': '	'}, None),
    ({'name': 'test', 'description': '   Test description.	'}, 'Test description.')
])
def test_trim_whitespace(thing_model, data_values, description_model_value):
    thing = thing_model(**data_values)
    assert thing.description == description_model_value


@pytest.mark.parametrize('data_values', [
    ({'name': '', 'description': 'Test description.'}),
    ({'name': '  ', 'description': 'Test description.'}),
    ({'name': '	', 'description': 'Test description.'})
])
def test_exception_on_whitespace(thing_model, data_values):
    with pytest.raises(ValidationError):
        thing_model(**data_values)



import inspect
import hydrothings.components as core_components
import hydrothings.schemas as core_schemas
from typing import Any, Callable
from pydantic import BaseModel, Field


def nested_entities_check(value: Any, field: Field) -> Any:
    """
    Validation for nested components.

    Runs validation on nested components in request bodies to avoid circular relationships in the API documentation.

    Parameters
    ----------
    value : Any
        The input value.
    field : Field
        The field being validated.

    Returns
    -------
    Any
        The output value.
    """

    nested_class_name = field.field_info.extra.get('nested_class')

    if nested_class_name:
        nested_class = getattr(core_components, nested_class_name)

        if isinstance(value, list):
            value = [
                nested_class(**sub_value.dict()) if isinstance(sub_value, core_schemas.NestedEntity) else sub_value
                for sub_value in value
            ]

        elif isinstance(value, core_schemas.NestedEntity):
            value = nested_class(**value.dict())

    return value


def whitespace_to_none(value: Any) -> Any:
    """
    Validation for whitespace values.

    Checks if string values are blank or only whitespace and converts them to NoneType.

    Parameters
    ----------
    value : Any
        The input value

    Returns
    -------
    Any
        The output value.
    """

    if isinstance(value, str):
        if value == '' or value.isspace():
            value = None
        else:
            value = value.strip()

    return value


def allow_partial(*fields) -> Callable:
    """
    Declares fields of a Pydantic model as optional.

    This decorator can be used to update required fields of a Pydantic model to be optional. This can be used in
    conjunction with exclude_unset to allow for missing fields, but still adhere to not_null requirements, such as when
    partially updating a datastore.

    :param fields: All fields which should be made optional. If none are passed, all fields will be made optional.
    :return dec: The optional class decorator.

    Parameters
    ----------
    fields : List[Field]
        All fields which should be made optional. If none are passed, all fields will be made optional.

    Returns
    -------
    Callable
        A class decorator that updates the required flag of selected fields to False.
    """

    def dec(_cls):
        for field in fields:
            _cls.__fields__[field].required = False
        return _cls

    if fields and inspect.isclass(fields[0]) and issubclass(fields[0], BaseModel):
        cls = fields[0]
        fields = cls.__fields__
        return dec(cls)

    return dec



from itertools import groupby
from typing import Union, List
from hydrothings.engine import SensorThingsRequest
from hydrothings.schemas import EntityId
from .schemas import ObservationPostBody, ObservationDataArray

fields = [
    ('id', 'id',),
    ('phenomenon_time', 'phenomenonTime',),
    ('result_time', 'resultTime',),
    ('result', 'result',),
    ('result_quality', 'resultQuality',),
    ('valid_time', 'validTime',),
    ('parameters', 'parameters',),
    ('feature_of_interest', 'FeatureOfInterest/id',)
]


def convert_to_data_array(
        request: SensorThingsRequest,
        response: dict,
        select: Union[list, None] = None
) -> dict:
    """
    Converts an Observations response dictionary to the dataArray format.

    Parameters
    ----------
    request : SensorThingsRequest
        The SensorThingsRequest object associated with the response.
    response : dict
        A SensorThings response dictionary.
    select
        A list of fields that should be included in the response.

    Returns
    -------
    dict
        A SensorThings response dictionary formatted as a dataArray.
    """

    if select:
        selected_fields = [
            field for field in fields if field[0] in select
        ]
    else:
        selected_fields = [
            field for field in fields if field[0] in ['result_time', 'result']
        ]

    datastream_url_template = f'{request.scheme}://{request.get_host()}{request.path[:-12]}Datastreams'

    response['value'] = [
        {
            'datastream': f'{datastream_url_template}({datastream_id})',
            'components': [
                field[1] for field in selected_fields
            ],
            'data_array': [
                [
                    observation[field[0]] for field in selected_fields
                ] for observation in observations
            ]
        } for datastream_id, observations in groupby(response['value'], key=lambda x: x['datastream_id'])
    ]

    return response


def parse_data_array(
        observation: List[ObservationDataArray]
) -> List[ObservationPostBody]:
    """
    Parses an ObservationDataArray object.

    Converts an ObservationDataArray object to a list of ObservationPostBody objects that can be loaded by the
    SensorThings engine.

    Parameters
    ----------
    observation: ObservationDataArray
        An ObservationDataArray object.

    Returns
    -------
    List[ObservationPostBody]
        A list of ObservationPostBody objects.
    """

    observations = []

    for datastream in observation:
        datastream_fields = [
            (field[0], field[1], get_field_index(datastream.components, field[1]),) for field in fields
        ]

        observations.extend([
            ObservationPostBody(
                datastream=datastream.datastream,
                **{
                    datastream_field[0]: entity[datastream_field[2]]
                    if datastream_field[0] != 'feature_of_interest'
                    else EntityId(
                        id=entity[datastream_field[2]]
                    )
                    for datastream_field in datastream_fields if datastream_field[2] is not None
                }
            ) for entity in datastream.data_array
        ])

    return observations


def get_field_index(components, field):
    """"""

    try:
        return components.index(field)
    except ValueError:
        return None
